
//
// Copyright (C) 2018 David Carpenter
// 
// Permission is hereby granted, free of charge, 
// to any person obtaining a copy of this software 
// and associated documentation files (the "Software"), 
// to deal in the Software without restriction, 
// including without limitation the rights to use, 
// copy, modify, merge, publish, distribute, sublicense, 
// and/or sell copies of the Software, and to permit 
// persons to whom the Software is furnished to do so, 
// subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be 
// included in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES 
// OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, 
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//

// compiler options
#pragma warning(disable : 4311) // type cast void * to unsigned int
#pragma warning(disable : 4312) // type cast unsigned int to void *
#pragma warning(disable : 4244) // warning C4244: 'argument' : conversion from 'LONG_PTR' to 'LONG', possible loss of data
#pragma warning(disable : 4996) // deprecation
#pragma warning(disable : 4201) // nonstandard extension used : nameless struct/union

#include <windows.h>
#include <Windowsx.h> // DeleteRgn
#include <commctrl.h>
#include <stdio.h>
#include <shlobj.h>
#include <strsafe.h>
#include <shlwapi.h>
#include "resource.h"

#define _NEW_SHIT_ID_ADD_LOCAL		10001
#define _NEW_SHIT_ID_DEL_LOCAL		10002
#define _NEW_SHIT_ID_CLEAR_LOCAL	10003

#define _NEW_SHIT_ID_ADD_REMOTE		10004
#define _NEW_SHIT_ID_DEL_REMOTE		10005
#define _NEW_SHIT_ID_CLEAR_REMOTE	10006

#define _NEW_SHIT_ID_LOCAL_LISTBOX	10007
#define _NEW_SHIT_ID_REMOTE_LISTBOX	10008
#define _NEW_SHIT_ID_A_MATCH_LISTBOX	10009
#define _NEW_SHIT_ID_B_MATCH_LISTBOX	10010

//#define _NEW_SHIT_ID_update_compare		10010

#define _NEW_SHIT_ID_RCOPY_TO_FOLDER	10020
#define _NEW_SHIT_ID_LCOPY_TO_FOLDER	10021
#define _NEW_SHIT_ID_RDELETE		10027
#define _NEW_SHIT_ID_LDELETE		10028

HWND local_listbox;
HWND remote_listbox;
HWND a_match_listbox;
HWND b_match_listbox;
HWND status_hwnd;
HWND a_static_hwnd;
HWND b_static_hwnd;
WNDPROC def_listbox_proc;

wchar_t local_location[MAX_PATH] = L"";
wchar_t remote_location[MAX_PATH] = L"";

wchar_t copyto_location[MAX_PATH] = L"";
HFONT bold_font;

typedef struct file_s
{
	wchar_t *path;
	wchar_t *short_path;
	struct file_s *next;
	unsigned __int64 size;
	struct file_s *match;
	struct file_s *left;
	int ismatch;
	int isremote;
	
}file_t;

file_t *a_file_start = 0;
file_t *local_file_last = 0;
file_t *b_file_start = 0;
file_t *remote_file_last = 0;

void debug_printf(const wchar_t *format,...)
{
	if (GetStdHandle(STD_OUTPUT_HANDLE) != INVALID_HANDLE_VALUE) 
	{
		va_list argptr;
		wchar_t wbuf[4096];
		DWORD NumberOfBytesWritten;

		va_start(argptr,format);
	
		vswprintf(wbuf,format,argptr);
		
		WriteConsoleW(GetStdHandle(STD_OUTPUT_HANDLE),wbuf,(int)wcslen(wbuf),&NumberOfBytesWritten,0);

		va_end(argptr);
	}
}

void os_draw_focus_rect(HDC hdc,RECT *prect,RECT *pcliprect,int xofs,int yofs)
{
	int x;
	int y;
	int run;
//	COLORREF pixel;

	if (!pcliprect) pcliprect = prect;

	// check valid range
	if (prect->left >= prect->right) return;
	if (prect->top >= prect->bottom) return;

	// completely invisible:
	if (prect->left > pcliprect->right) return;
	if (prect->top > pcliprect->bottom) return;
	if (prect->right < pcliprect->left) return;
	if (prect->bottom < pcliprect->top) return;
	
	{
		HBRUSH hbrush;
		HBITMAP hbitmap;
		BITMAP bitmap;
		HGDIOBJ lasthbrush;
		COLORREF lasttextcolor;
		COLORREF lastbkcolor;
		int lastbkmode;
		
		unsigned int pixels[8] = 
		{
			0xAAAAAAAA,
			0x55555555,
			0xAAAAAAAA,
			0x55555555,
			0xAAAAAAAA,
			0x55555555,
			0xAAAAAAAA,
			0x55555555,
		};
		
		bitmap.bmType = 0;
		bitmap.bmWidth = 8;
		bitmap.bmHeight = 8;
		bitmap.bmWidthBytes = 4;
		bitmap.bmPlanes = 1;
		bitmap.bmBits = (LPVOID)pixels;
		bitmap.bmBitsPixel = 1;

		hbitmap = CreateBitmapIndirect(&bitmap);
		if (hbitmap)
		{
		
			hbrush = CreatePatternBrush(hbitmap);
			if (hbrush)
			{
				SetBrushOrgEx(hdc,xofs,yofs,0);
				lasthbrush = SelectObject(hdc,hbrush);
				
				lastbkmode = SetBkMode(hdc,TRANSPARENT);
				lastbkcolor = SetBkColor(hdc,RGB(0,0,0));
				lasttextcolor = SetTextColor(hdc,RGB(255,255,255));

//				PatBlt(hdc,prect->left,prect->top,prect->right -prect->left,prect->bottom -prect->top,PATINVERT);

/////////////////

				if (prect->top >= pcliprect->top)
				{
					x = prect->left;
					run = prect->right;
					if (x < pcliprect->left) x = pcliprect->left;
					if (run > pcliprect->right) run = pcliprect->right;

//					if ((x + prect->top+xofs+yofs) & 1) x++;
					if (run>x) PatBlt(hdc,x,prect->top,run-x,1,PATINVERT);
				}

				if (prect->bottom-1 <= pcliprect->bottom-1)
				{
					if (prect->top != prect->bottom-1)
					{
						x = prect->left;
						run = prect->right;
						if (x < pcliprect->left) x = pcliprect->left;
						if (run > pcliprect->right) run = pcliprect->right;

//						if ((x + prect->bottom-1+xofs+yofs) & 1) x++;
						if (run>x) PatBlt(hdc,x,prect->bottom-1,run-x,1,PATINVERT);
					}
				}

				if (prect->left >= pcliprect->left)
				{
					y = prect->top+1;
					run = prect->bottom-1;
					if (y < pcliprect->top) y = pcliprect->top;
					if (run > pcliprect->bottom) run = pcliprect->bottom;

//					if ((y + prect->left+xofs+yofs) & 1) y++;
					if (run>y) PatBlt(hdc,prect->left,y,1,run-y,PATINVERT);
				}

				if (prect->right-1 <= pcliprect->right-1)
				{
					if (prect->left != prect->right-1)
					{
						y = prect->top+1;
						run = prect->bottom-1;
						if (y < pcliprect->top) y = pcliprect->top;
						if (run > pcliprect->bottom) run = pcliprect->bottom;

//						if ((y + prect->right-1+xofs+yofs) & 1) y++;
						if (run>y) PatBlt(hdc,prect->right-1,y,1,run-y,PATINVERT);
					}
				}
				
/////////////////

				SetBkMode(hdc,lastbkmode);
				SetBkColor(hdc,lastbkcolor);
				SetTextColor(hdc,lasttextcolor);
				
				SelectObject(hdc,lasthbrush);
				DeleteBrush(hbrush);
			}
			
			DeleteBitmap(hbitmap);
		}
		
		return ;
	}
/*
	if (!pcliprect) pcliprect = prect;

	// check valid range
	if (prect->left >= prect->right) return;
	if (prect->top >= prect->bottom) return;

	// completely invisible:
	if (prect->left > pcliprect->right) return;
	if (prect->top > pcliprect->bottom) return;
	if (prect->right < pcliprect->left) return;
	if (prect->bottom < pcliprect->top) return;
	
	if (prect->top >= pcliprect->top)
	{
		x = prect->left;
		run = prect->right;
		if (x < pcliprect->left) x = pcliprect->left;
		if (run > pcliprect->right) run = pcliprect->right;

		if ((x + prect->top+xofs+yofs) & 1) x++;
		for(;x < run;x+=2)
		{	
			pixel = GetPixel(hdc,x,prect->top);
			SetPixelV(hdc,x,prect->top,~(pixel|0xff000000));
		}
	}
	
	if (prect->bottom-1 <= pcliprect->bottom-1)
	{
		if (prect->top != prect->bottom-1)
		{
			x = prect->left;
			run = prect->right;
			if (x < pcliprect->left) x = pcliprect->left;
			if (run > pcliprect->right) run = pcliprect->right;

			if ((x + prect->bottom-1+xofs+yofs) & 1) x++;
			for(;x < run;x+=2)
			{	
				pixel = GetPixel(hdc,x,prect->bottom-1);
				SetPixelV(hdc,x,prect->bottom-1,~(pixel|0xff000000));
			}
		}
	}

	if (prect->left >= pcliprect->left)
	{
		y = prect->top+1;
		run = prect->bottom-1;
		if (y < pcliprect->top) y = pcliprect->top;
		if (run > pcliprect->bottom) run = pcliprect->bottom;

		if ((y + prect->left+xofs+yofs) & 1) y++;
		
		for(;y < run;y+=2)
		{	
			pixel = GetPixel(hdc,prect->left,y);
			SetPixelV(hdc,prect->left,y,~(pixel|0xff000000));
		}
	}

	if (prect->right-1 <= pcliprect->right-1)
	{
		if (prect->left != prect->right-1)
		{
			y = prect->top+1;
			run = prect->bottom-1;
			if (y < pcliprect->top) y = pcliprect->top;
			if (run > pcliprect->bottom) run = pcliprect->bottom;

			if ((y + prect->right-1+xofs+yofs) & 1) y++;
			
			for(;y < run;y+=2)
			{	
				pixel = GetPixel(hdc,prect->right-1,y);
				SetPixelV(hdc,prect->right-1,y,~(pixel|0xff000000));
			}
		}
	}
	*/
}

file_t *add_file(const wchar_t *path,const wchar_t *short_path,file_t **start,file_t **last,unsigned __int64 size)
{
	file_t *f;
	
//	debug_printf(L"add_file: %s\n",short_path);
	
	f = (file_t *)malloc(sizeof(file_t));
	
	f->path = (wchar_t *)malloc((wcslen(path)+1)*2);
	CopyMemory(f->path,path,(wcslen(path)+1)*2);

	f->short_path = (wchar_t *)malloc((wcslen(short_path)+1)*2);
	CopyMemory(f->short_path,short_path,(wcslen(short_path)+1)*2);

	f->size = size;
	
	if (*start)
	{
		(*last)->next = f;
	}
	else
	{
		*start = f;
	}
	
	f->next = 0;
	*last = f;
	
	return f;
}

void scan(wchar_t *path,const wchar_t *short_path,file_t **start,file_t **last)
{
	WIN32_FIND_DATAW fd;
	wchar_t buf[1024];
	wchar_t shortbuf[1024];
	HANDLE h;
	
	swprintf(buf,L"%s\\*.*",path);
	
	h = FindFirstFileW(buf,&fd);
	
	if (h != INVALID_HANDLE_VALUE)
	{
		for(;;)
		{
			if (fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
			{
				if (wcscmp(fd.cFileName,L".") == 0)
				{
				}
				else
				if (wcscmp(fd.cFileName,L"..") == 0)
				{
				}
				else
				{
					swprintf(buf,L"%s\\%s",path,fd.cFileName);
					
					if (*short_path)
					{
						swprintf(shortbuf,L"%s\\%s",short_path,fd.cFileName);
					}
					else
					{
						swprintf(shortbuf,L"%s",fd.cFileName);
					}
					
					scan(buf,shortbuf,start,last);
				}
			}
			
			if (!FindNextFileW(h,&fd)) break;
		}
	
		FindClose(h);
	}

	
	swprintf(buf,L"%s\\*.*",path);
	
	h = FindFirstFileW(buf,&fd);
	
	if (h != INVALID_HANDLE_VALUE)
	{
		for(;;)
		{
			if (!(fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
			{
				swprintf(buf,L"%s\\%s",path,fd.cFileName);

				if (*short_path)
				{
					swprintf(shortbuf,L"%s\\%s",short_path,fd.cFileName);
				}
				else
				{
					swprintf(shortbuf,L"%s",fd.cFileName);
				}
				
				add_file(buf,shortbuf,start,last,(((unsigned __int64)fd.nFileSizeHigh) << 32) | fd.nFileSizeLow);
			}
			
			if (!FindNextFileW(h,&fd)) break;
		}
	
		FindClose(h);
	}
}

file_t *find_file(file_t *start,wchar_t *shortpath)
{
	file_t *f;
	
	f = start;
	while(f)
	{
		if (wcsicmp(f->short_path,shortpath) == 0) return f;
		
		f = f->next;
	}
	
	return 0;
}


void os_dlg_setredraw(HWND hwnd,BOOL redraw)
{
	SetWindowRedraw(hwnd,redraw);
	
	if (redraw)
	{
		InvalidateRect(hwnd,0,FALSE);
	}
}

void update_compare(void)
{
	file_t *f;
	file_t *f2;

	os_dlg_setredraw(a_match_listbox,FALSE);
	os_dlg_setredraw(b_match_listbox,FALSE);

	ListBox_ResetContent(a_match_listbox);
	ListBox_ResetContent(b_match_listbox);
	
	// a
	f = a_file_start;
	while(f)
	{
		if (!find_file(b_file_start,f->short_path))
		{
			int index;
			
			index = ListBox_AddString(a_match_listbox,f->short_path);

			SendMessage(a_match_listbox,LB_SETITEMDATA,index,(LPARAM)f); 
		}

		f = f->next;
	}
	
	// b
	f = b_file_start;
	while(f)
	{
		if (!find_file(a_file_start,f->short_path))
		{
			int index;
			
			index = ListBox_AddString(b_match_listbox,f->short_path);

			SendMessage(a_match_listbox,LB_SETITEMDATA,index,(LPARAM)f); 
		}

		f = f->next;
	}
	
	os_dlg_setredraw(a_match_listbox,TRUE);
	os_dlg_setredraw(b_match_listbox,TRUE);
	
}

int CALLBACK _os_BrowseCallbackProc(HWND hwnd,UINT uMsg,LPARAM lParam,LPARAM lpData)
{
	switch(uMsg)
	{
		case BFFM_INITIALIZED:
			SendMessage(hwnd,BFFM_SETSELECTIONW,(WPARAM)1,(LPARAM)lpData);
			break;
	}
	
	return 0;
}

//FIXME: resolve shortcuts..
int os_browse_for_folder(HWND parent,const wchar_t *title,wchar_t *buf)
{
	BROWSEINFOW bi;
	ITEMIDLIST *iil;

	ZeroMemory(&bi,sizeof(bi));
	bi.hwndOwner = parent;
	bi.lpszTitle = title;
	bi.pszDisplayName = buf;
	bi.ulFlags = BIF_USENEWUI;
	bi.lParam = (LPARAM)buf;
	bi.lpfn = _os_BrowseCallbackProc;
	
	iil = SHBrowseForFolderW(&bi);
	if (iil)
	{
		SHGetPathFromIDListW(iil,buf);

		CoTaskMemFree(iil);

		return 1;
	}
	
	return 0;
}

void os_set_window_id(HWND hwnd,int id) 
{
	SetWindowLong(hwnd,GWL_ID,(LONG)id);
}

HWND os_create_button(HWND parent,int id,const wchar_t *text)
{
	HWND hwnd;
	
	hwnd = CreateWindowExW(
		WS_EX_NOPARENTNOTIFY,
		WC_BUTTONW,
		text,
		WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_CHILDWINDOW | WS_VISIBLE | WS_TABSTOP,
		0,0,0,0,
		parent,0,GetModuleHandle(0),0);
		
	os_set_window_id(hwnd,id);
	
	SendMessage(hwnd,WM_SETFONT,(WPARAM)(HFONT)GetStockObject(DEFAULT_GUI_FONT),0);
//	BringWindowToTop(hwnd);
//	SetFocus(hwnd);
//	SetForegroundWindow(hwnd);
	
	return hwnd;
}

// create a static window
HWND os_create_static(HWND parent,const wchar_t *text,int x,int y)
{
	HWND hwnd;
	int wide;
	HDC hdc;
	HGDIOBJ lastfont;
	SIZE size;

	hdc = GetDC(parent);
	
	lastfont = SelectObject(hdc,GetStockObject(DEFAULT_GUI_FONT));
	
	GetTextExtentPointW(hdc,text,wcslen(text),&size);
	
	wide = size.cx;

	SelectObject(hdc,lastfont);
	
	ReleaseDC(parent,hdc);

	// MUST be a single line

	hwnd = CreateWindowExW(
		0,
		WC_STATICW,
		text,
		WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_CHILDWINDOW | WS_VISIBLE,
		x,y,wide,13,
		parent,0,GetModuleHandle(0),0);

	
	SendMessage(hwnd,WM_SETFONT,(WPARAM)(HFONT)GetStockObject(DEFAULT_GUI_FONT),0);
//HACKHACKHACK:
//	SetClassLong(hwnd,GCL_HBRBACKGROUND,(long)CreateSolidBrush(RGB(0,0,0)));
//	BringWindowToTop(hwnd);

	
	return hwnd;
}

HWND os_create_list_box(HWND parent,int id,int x,int y,int wide,int high,int sort,int group,int custom)
{
	HWND hwnd;
	
	hwnd = CreateWindowExW(
		WS_EX_CLIENTEDGE,
		WC_LISTBOXW,
		L"",
		WS_VISIBLE | WS_CLIPCHILDREN | WS_CLIPSIBLINGS | WS_CHILD | WS_VSCROLL | WS_HSCROLL | WS_TABSTOP | (group?WS_GROUP:0) | LBS_NOTIFY | LBS_NOINTEGRALHEIGHT | (sort?LBS_SORT:0) | (custom?(LBS_OWNERDRAWFIXED):0) | LBS_EXTENDEDSEL,
		x,y,wide,high,
		parent,0,GetModuleHandle(0),0);
		
	SendMessage(hwnd,WM_SETFONT,(WPARAM)(HFONT)GetStockObject(DEFAULT_GUI_FONT),0);
	
	os_set_window_id(hwnd,id);
	
	return hwnd;
}

void os_client_to_window_rect(HWND hwnd,RECT *window_rect,int window_x,int window_y,int client_wide,int client_high)
{
	window_rect->left = 0;
	window_rect->top = 0;
	window_rect->right = client_wide;
	window_rect->bottom = client_high;
	
	AdjustWindowRect(window_rect,GetWindowStyle(hwnd),0);
	
	
	window_rect->right -= window_rect->left;
	window_rect->bottom -= window_rect->top;
	window_rect->left = window_x;
	window_rect->top = window_y;
	window_rect->right += window_x;
	window_rect->bottom += window_y;

}

// make a rect fully visible
// pushes the rect onto the screen, the same as menus.
void os_make_rect_fully_visible(RECT *prect,int workarea)
{
	HMONITOR hmonitor;
	MONITORINFO mi;
	int wide;
	int high;
	RECT *workrect;

	// we hit something so reposition the window on this monitor.
	// note that MonitorFromRect can return null, even when MONITOR_DEFAULTTOPRIMARY is specified.
	hmonitor = MonitorFromRect(prect,MONITOR_DEFAULTTOPRIMARY);
	if (!hmonitor)
	{
		// just leave it, rather than fail.
		return;
	}

	mi.cbSize = sizeof(mi);
	GetMonitorInfo(hmonitor,&mi);
	
	if (workarea)
	{
		workrect = &mi.rcWork;
	}
	else
	{
		workrect = &mi.rcMonitor;
	}
	
	// get window width and height,
	wide = prect->right - prect->left;
	high = prect->bottom - prect->top;

	// push the window 
	if (prect->right > workrect->right) 
	{
		prect->left = workrect->right - wide;
		prect->right = workrect->right;
	}
	
	if (prect->bottom > workrect->bottom) 
	{
		prect->top = workrect->bottom - high;
		prect->bottom = workrect->bottom;
	}

	// push the window 
	if (prect->left < workrect->left) 
	{
		prect->left = workrect->left;
		prect->right = workrect->left + wide;
	}
	
	if (prect->top < workrect->top) 
	{
		prect->top = mi.rcMonitor.top;
		prect->bottom = workrect->top + high;
	}
}

void os_set_window_rect(HWND hwnd,int x,int y,int wide,int high)
{
	SetWindowPos(hwnd,0,x,y,wide,high,SWP_NOZORDER|SWP_NOACTIVATE);
}

void os_center_dialog(HWND parent,HWND hwnd,int wide,int high)
{
	RECT rect;
	int x;
	int y;

	// position window at center of the everything window
	GetWindowRect(parent,&rect);
	x = rect.left + ((rect.right - rect.left) / 2);
	y = rect.top + ((rect.bottom - rect.top) / 2);

	os_client_to_window_rect(
		hwnd,
		&rect,
		0,
		0,
		wide,
		high);
		
	rect.right -= rect.left;
	rect.bottom -= rect.top;
	
	x -= (rect.right / 2);
	y -= (rect.bottom / 2);

	rect.left = x;
	rect.top = y;
	rect.right += x;
	rect.bottom += y;

	os_make_rect_fully_visible(&rect,1);
	os_set_window_rect(hwnd,rect.left,rect.top,rect.right-rect.left,rect.bottom - rect.top);
}

INT_PTR __stdcall a_match_listbox_proc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{
	switch(msg)
	{
		case 0x020A:
		{
			CallWindowProc(def_listbox_proc,hwnd,msg,wParam,lParam);
			CallWindowProc(def_listbox_proc,b_match_listbox,msg,wParam,lParam);

			return 0;
		}
			
		case WM_VSCROLL:
		{
			CallWindowProc(def_listbox_proc,hwnd,msg,wParam,lParam);
			CallWindowProc(def_listbox_proc,b_match_listbox,msg,wParam,lParam);
			
			return 0;
		}
	}
	
	return CallWindowProc(def_listbox_proc,hwnd,msg,wParam,lParam);
}

INT_PTR __stdcall b_match_listbox_proc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{
	switch(msg)
	{
		case 0x020A:
		{
			CallWindowProc(def_listbox_proc,hwnd,msg,wParam,lParam);
			CallWindowProc(def_listbox_proc,a_match_listbox,msg,wParam,lParam);

			return 0;
		}
			
		case WM_VSCROLL:
		{
			CallWindowProc(def_listbox_proc,hwnd,msg,wParam,lParam);
			CallWindowProc(def_listbox_proc,a_match_listbox,msg,wParam,lParam);
			
			return 0;
		}
	}
	
	return CallWindowProc(def_listbox_proc,hwnd,msg,wParam,lParam);
}

void on_size(HWND hwnd)
{
	int x;
	int y;
	int wide;
	int high;
	RECT rect;
	int listhigh1;
	int listhigh2;
	
	GetClientRect(hwnd,&rect);
	
	// create buttons
	/*
	os_create_button(hwnd,ABOUT_ID_CREDITS,1,1,localization_strings[LOCALIZATION_CREDITS2],x + wide + 6 - ((75 + 6) * 2) - 6,y + high - 23 - 7,75,23);
	os_create_button(hwnd,IDOK,1,1,localization_strings[LOCALIZATION_OK],x + wide + 6 - ((75 + 6) * 1) - 6,y + high - 23 - 7,75,23);
	os_set_default_button(hwnd,IDOK);
	
	utf8_printf(buf,sizeof(buf),localization_strings[LOCALIZATION_ABOUT_FORMAT],localization_strings[LOCALIZATION_EVERYTHING]);
	
	os_set_window_text(hwnd,buf);
	
	x += 48;
	wide -= 48;
	y += 102;
	
	os_create_edit_box(hwnd,ABOUT_ID_EVERYTHING,localization_strings[LOCALIZATION_EVERYTHING],0,0,0,0,x,y,wide,16);
	SendDlgItemMessage(hwnd,ABOUT_ID_EVERYTHING,EM_SETREADONLY,(WPARAM)TRUE,0);
	
	y += 20;

	utf8_printf(buf,sizeof(buf),localization_strings[LOCALIZATION_VERSION_FORMAT],VERSION_STRING_RELEASE);
	os_create_edit_box(hwnd,ABOUT_ID_VERSION,buf,0,0,0,0,x,y,wide,16);
	SendDlgItemMessage(hwnd,ABOUT_ID_VERSION,EM_SETREADONLY,(WPARAM)TRUE,0);

	y += 20;

	os_create_edit_box(hwnd,ABOUT_ID_COPYRIGHT,localization_strings[LOCALIZATION_COPYRIGHT],0,0,0,0,x,y,wide,16);
	SendDlgItemMessage(hwnd,ABOUT_ID_COPYRIGHT,EM_SETREADONLY,(WPARAM)TRUE,0);

	// show the window
	ShowWindow(hwnd,SW_SHOW);
	SetForegroundWindow(hwnd);
	SetFocus(GetDlgItem(hwnd,ABOUT_ID_CREDITS));			
	*/
	
	x = 0;
	y = 0;
	wide = rect.right - rect.left;
	high = rect.bottom - rect.top;
	
	listhigh1 = ((high - 80 - 13 - 12 - 6) * 1) / 4;
	listhigh2 = ((high - 80 - 13 - 12 - 6) * 3) / 4;
	
//	y += 13 + 4;

	os_set_window_rect(a_static_hwnd,x+4,y+4,(wide-12)/2,12);
	os_set_window_rect(b_static_hwnd,x+4+4+(wide-12)/2,y+4,(wide-12)/2,12);

	y += 12 + 6;

	os_set_window_rect(local_listbox,x+4,y+4,(wide-12)/2,listhigh1);
	os_set_window_rect(remote_listbox,x+4+4+(wide-12)/2,y+4,(wide-12)/2,listhigh1);

	y += listhigh1 + 4;

	os_set_window_rect(GetDlgItem(hwnd,_NEW_SHIT_ID_ADD_LOCAL),x+4+((wide-12)/2)-75-4-75,y+4,75,23);
//	os_set_window_rect(GetDlgItem(hwnd,_NEW_SHIT_ID_DEL_LOCAL),x+4+((wide-12)/2)-75-4-75,y+4,75,23);
	os_set_window_rect(GetDlgItem(hwnd,_NEW_SHIT_ID_CLEAR_LOCAL),x+4+((wide-12)/2)-75,y+4,75,23);
	
	os_set_window_rect(GetDlgItem(hwnd,_NEW_SHIT_ID_ADD_REMOTE),x+4+(wide-12)/2+4+((wide-12)/2)-75-4-75,y+4,75,23);
//	os_set_window_rect(GetDlgItem(hwnd,_NEW_SHIT_ID_DEL_REMOTE),x+4+(wide-12)/2+4+((wide-12)/2)-75-4-75,y+4,75,23);
	os_set_window_rect(GetDlgItem(hwnd,_NEW_SHIT_ID_CLEAR_REMOTE),x+4+(wide-12)/2+4+((wide-12)/2)-75,y+4,75,23);
	
	y += 4 + 23;
	
//	os_set_window_rect(GetDlgItem(hwnd,_NEW_SHIT_ID_update_compare),x+((wide)/2)-((75)/2),y+4,75,23);
//	os_set_window_rect(GetDlgItem(hwnd,_NEW_SHIT_ID_update_compare),x+4,y+4,wide-8,23);
	
//	y += 4 + 23;

	os_set_window_rect(a_match_listbox,x+4,y+4,(wide-12)/2,listhigh2);
	os_set_window_rect(b_match_listbox,x+4+4+(wide-12)/2,y+4,(wide-12)/2,listhigh2);
	
	y += listhigh2 + 4;
	
	os_set_window_rect(GetDlgItem(hwnd,_NEW_SHIT_ID_LCOPY_TO_FOLDER),x+4+wide-100-4-4-((wide-12)/2)-4,y+4,100,23);
	os_set_window_rect(GetDlgItem(hwnd,_NEW_SHIT_ID_RCOPY_TO_FOLDER),x+4+wide-100-4-4,y+4,100,23);

	os_set_window_rect(GetDlgItem(hwnd,_NEW_SHIT_ID_LDELETE),x+4+wide-100-4-4-((wide-12)/2)-4-100-4,y+4,100,23);
	os_set_window_rect(GetDlgItem(hwnd,_NEW_SHIT_ID_RDELETE),x+4+wide-100-4-4-100-4,y+4,100,23);

	SendMessage(status_hwnd,WM_SIZE,0,0);
}

INT_PTR __stdcall _newshit_proc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{
	switch(msg)
	{
		case WM_GETMINMAXINFO:
		{
			MINMAXINFO *mmi = (MINMAXINFO *)lParam;
			
			mmi->ptMinTrackSize.x = 4 + 4 +4 + (75 + 4 + 75 + 4 + 75) + 4 + (75 + 4 + 75 + 4 + 75) + 4;
			mmi->ptMinTrackSize.y = 141;
			
			break;
		}
			
		case WM_INITDIALOG:
		{
			SetWindowTextW(hwnd,L"New Shit");

			SendMessage(hwnd,WM_SETICON,ICON_SMALL,(LPARAM)LoadIconW(GetModuleHandle(0),MAKEINTRESOURCE(IDI_ICON1)));
			SendMessage(hwnd,WM_SETICON,ICON_BIG,(LPARAM)LoadIconW(GetModuleHandle(0),MAKEINTRESOURCE(IDI_ICON2)));
			
			// position window at center of the everything window
			os_center_dialog(GetDesktopWindow(),hwnd,720,640);
			
			a_static_hwnd = os_create_static(hwnd,L"A:",0,0);
			b_static_hwnd = os_create_static(hwnd,L"B:",0,0);

			local_listbox = os_create_list_box(hwnd,_NEW_SHIT_ID_LOCAL_LISTBOX,0,0,0,0,1,0,0);
			remote_listbox = os_create_list_box(hwnd,_NEW_SHIT_ID_REMOTE_LISTBOX,0,0,0,0,1,0,0);
			
			os_create_button(hwnd,_NEW_SHIT_ID_ADD_LOCAL,L"&Add...");
//			os_create_button(hwnd,_NEW_SHIT_ID_DEL_LOCAL,L"&Remove");
			os_create_button(hwnd,_NEW_SHIT_ID_CLEAR_LOCAL,L"&Clear");

			os_create_button(hwnd,_NEW_SHIT_ID_ADD_REMOTE,L"&Add...");
//			os_create_button(hwnd,_NEW_SHIT_ID_DEL_REMOTE,L"&Remove");
			os_create_button(hwnd,_NEW_SHIT_ID_CLEAR_REMOTE,L"&Clear");

			a_match_listbox = os_create_list_box(hwnd,_NEW_SHIT_ID_A_MATCH_LISTBOX,0,0,0,0,0,0,0);
			b_match_listbox = os_create_list_box(hwnd,_NEW_SHIT_ID_B_MATCH_LISTBOX,0,0,0,0,0,0,0);
			
			os_create_button(hwnd,_NEW_SHIT_ID_LCOPY_TO_FOLDER,L"&Copy");
			os_create_button(hwnd,_NEW_SHIT_ID_RCOPY_TO_FOLDER,L"&Copy");

			os_create_button(hwnd,_NEW_SHIT_ID_LDELETE,L"&Delete");
			os_create_button(hwnd,_NEW_SHIT_ID_RDELETE,L"&Delete");

			status_hwnd = CreateWindowExW(0,STATUSCLASSNAMEW,L"",
				WS_CLIPCHILDREN | WS_CLIPSIBLINGS | WS_CHILD | SBARS_SIZEGRIP | WS_VISIBLE | SBT_TOOLTIPS,
				0,0,0,0,
				hwnd,0,GetModuleHandle(0),0);

//			os_create_button(hwnd,_NEW_SHIT_ID_update_compare,L"&update_compare");

			on_size(hwnd);
			
			DragAcceptFiles(hwnd,TRUE);
			
//			AllocConsole();

			break;
		}
		
		case WM_SIZE:
		
			on_size(hwnd);
			
			break;
			
		case WM_DESTROY: 
		
			DragAcceptFiles(hwnd,FALSE);
			
			PostQuitMessage(0);
			
			break;
			
		case WM_CLOSE:
			EndDialog(hwnd,0);
			break;

/*			
		case WM_CTLCOLORLISTBOX:
		
			return (LRESULT)CreateSolidBrush(RGB(rand(),rand(),rand()));
		
			break;
			*/
		case WM_COMMAND: 
			
			switch(LOWORD(wParam))
			{
				case _NEW_SHIT_ID_ADD_LOCAL:
				{
					if (os_browse_for_folder(hwnd,L"Browse for local folder",local_location))
					{
						int i;
						i = ListBox_InsertString(local_listbox,-1,local_location);
						
						ListBox_SetCurSel(local_listbox,i);
						
						scan(local_location,L"",&a_file_start,&local_file_last);

						update_compare();
					}

					break;
				}
					
				case _NEW_SHIT_ID_DEL_LOCAL:
				{
					int i;
					
					i = ListBox_GetCurSel(local_listbox);
					if (i != LB_ERR)
					{
						ListBox_DeleteString(local_listbox,i);
					}

					update_compare();

					break;
				}
					
				case _NEW_SHIT_ID_CLEAR_LOCAL:
				{
					ListBox_ResetContent(local_listbox);

					a_file_start = 0;
					local_file_last = 0;
					
					update_compare();

					break;
				}
					
				case _NEW_SHIT_ID_ADD_REMOTE:
				{
					if (os_browse_for_folder(hwnd,L"Browse for remote folder",remote_location))
					{
						int i;
						i = ListBox_InsertString(remote_listbox,-1,remote_location);
						
						ListBox_SetCurSel(remote_listbox,i);

						scan(remote_location,L"",&b_file_start,&remote_file_last);

						update_compare();
					}
					
					break;
				}
					
				case _NEW_SHIT_ID_DEL_REMOTE:
				{
					int i;
					
					i = ListBox_GetCurSel(remote_listbox);
					if (i != LB_ERR)
					{
						ListBox_DeleteString(remote_listbox,i);
					}

					update_compare();

					break;
				}
					
				case _NEW_SHIT_ID_CLEAR_REMOTE:
				{
					ListBox_ResetContent(remote_listbox);

					b_file_start = 0;
					remote_file_last = 0;

					update_compare();

					break;
				}
					
				case _NEW_SHIT_ID_LCOPY_TO_FOLDER:
				case _NEW_SHIT_ID_RCOPY_TO_FOLDER:
				{
					HWND listid;
						
					if (LOWORD(wParam) == _NEW_SHIT_ID_RCOPY_TO_FOLDER)
					{
						listid = b_match_listbox;
					}
					else
					{
						listid = a_match_listbox;
					}

					if (ListBox_GetSelCount(listid))
					{

						if (os_browse_for_folder(hwnd,L"Browse for the copy destination location",copyto_location))
						{
							int i;
							SHFILEOPSTRUCT fo;
							wchar_t *tobuf;
							int fromsize;
							int tosize;
							wchar_t *frombuf;
							wchar_t *p;
							wchar_t *d;
							wchar_t buf[1024];
							
							fromsize = 0;
							tosize = 0;

							// right is remote.
							for(i=0;i<ListBox_GetCount(listid);i++)
							{
								if (ListBox_GetSel(listid,i))
								{
									file_t *f;
									
									f = (file_t *)SendMessage(listid,LB_GETITEMDATA,i,0); 

									fromsize += wcslen(f->path) + 1;
										
									wsprintf(buf,L"%s\\%s",copyto_location,f->short_path);

	debug_printf(L"%s => %s\n",f->path,buf);

									tosize += wcslen(buf) + 1;
								}
							}
							
							fromsize++;
							tosize++;
							
							frombuf = (wchar_t *)malloc(fromsize * 2);
							p = frombuf;

							tobuf = (wchar_t *)malloc(tosize * 2);
							d = tobuf;
							
							for(i=0;i<ListBox_GetCount(listid);i++)
							{
								if (ListBox_GetSel(listid,i))
								{
									file_t *f;
									
									f = (file_t *)SendMessage(listid,LB_GETITEMDATA,i,0); 

									wcscpy(p,f->path);
									p += wcslen(f->path) + 1;

									wsprintf(d,L"%s\\%s",copyto_location,f->short_path);
									d += wcslen(d) + 1;
								}
							}
							
							*p = 0;
							*d = 0;
							
							ZeroMemory(&fo,sizeof(fo));
							fo.pTo = tobuf;
							fo.pFrom = frombuf;
							fo.fFlags = FOF_ALLOWUNDO | FOF_MULTIDESTFILES | FOF_NOCONFIRMMKDIR;
							fo.hwnd = hwnd;
							fo.wFunc = FO_COPY;

							// weeeeeeeeeeeeeee									
							SHFileOperation(&fo);
						}
					}
						

					break;
				}

				case _NEW_SHIT_ID_LDELETE:
				case _NEW_SHIT_ID_RDELETE:
				{
					int i;
					file_t *f;
					SHFILEOPSTRUCT fo;
					int fromsize;
					wchar_t *frombuf;
					wchar_t *p;
					wchar_t *d;
					wchar_t buf[1024];
					HWND listid;
					
					fromsize = 0;
					
					if (LOWORD(wParam) == _NEW_SHIT_ID_RDELETE)
					{
						listid = b_match_listbox;
					}
					else
					{
						listid = a_match_listbox;
					}

					// right is remote.
					for(i=0;i<ListBox_GetCount(listid);i++)
					{
						if (ListBox_GetSel(listid,i))
						{
							f = (file_t *)SendMessage(listid,LB_GETITEMDATA,i,0); 

							if (f)
							{
								fromsize += wcslen(f->path) + 1;
								
debug_printf(L"DELETE %s\n",f->path);
							}
						}
					}
					
					fromsize++;
					
					frombuf = (wchar_t *)malloc(fromsize * 2);
					p = frombuf;

					for(i=0;i<ListBox_GetCount(listid);i++)
					{
						if (ListBox_GetSel(listid,i))
						{
							f = (file_t *)SendMessage(listid,LB_GETITEMDATA,i,0); 

							if (f)
							{
								wcscpy(p,f->path);
								p += wcslen(f->path) + 1;
							}
						}
					}
					
					*p = 0;
					
					ZeroMemory(&fo,sizeof(fo));
					fo.pTo = 0;
					fo.pFrom = frombuf;
					fo.fFlags = FOF_MULTIDESTFILES;
					if (!(GetKeyState(VK_SHIFT) & 0x80))
					{
						fo.fFlags |= FOF_ALLOWUNDO;
					}
					fo.hwnd = hwnd;
					fo.wFunc = FO_DELETE;

					// weeeeeeeeeeeeeee									
					SHFileOperation(&fo);
					
					break;
				}
/*					
				case _NEW_SHIT_ID_update_compare:
				{
					update_compare();
					break;
				}
*/					
				case IDOK:
					EndDialog(hwnd,0);
					break;	

				case IDCANCEL:
					EndDialog(hwnd,0);
					break;	
			}
			 
			break;
			
		case WM_DROPFILES:
		{
			int files,i;
			wchar_t buf[1024];
			POINT pt;
			HWND listhwnd;
			
			DragQueryPoint((HDROP)wParam,&pt); 
			
			ClientToScreen(hwnd,&pt);
			
			listhwnd = WindowFromPoint(pt);
			
			if ((listhwnd == local_listbox) || (listhwnd == remote_listbox))
			{
				files = DragQueryFileW((HDROP)wParam,0xffffffff,buf,1024);
				
				for(i=0;i<files;i++)
				{
					DragQueryFileW((HDROP)wParam,i,buf,1024);

					ListBox_InsertString(listhwnd,-1,buf);
				}
			}
			
			break;
		}
		
	}
	
	return FALSE;
}

void os_create_blank_dialog(HWND parent,DLGPROC proc,void *param)
{
    HGLOBAL hgbl;
    
    struct blank_dialog_s
    {
		DWORD style;
		DWORD dwExtendedStyle;
		WORD cdit;
		short x;
		short y;
		short cx;
		short cy;
		short menu;
		short predefined_dialog_class;
		short name;
    }*blank_dialog;
    
    hgbl = GlobalAlloc(GMEM_ZEROINIT,sizeof(struct blank_dialog_s));
    blank_dialog = (struct blank_dialog_s *)GlobalLock(hgbl);
 
    // Define a dialog box.
    blank_dialog->style = WS_BORDER | WS_SYSMENU | DS_3DLOOK | DS_MODALFRAME | WS_CAPTION | WS_CLIPCHILDREN | WS_CLIPSIBLINGS | WS_VISIBLE | WS_THICKFRAME | WS_SYSMENU | WS_MAXIMIZEBOX | WS_MINIMIZEBOX;
	blank_dialog->dwExtendedStyle = WS_EX_DLGMODALFRAME | WS_EX_CONTROLPARENT;
    blank_dialog->cdit = 0;
    blank_dialog->x = 0;
    blank_dialog->y = 0;
    blank_dialog->cx = 0; 
    blank_dialog->cy = 0;
    blank_dialog->menu = 0;
    blank_dialog->predefined_dialog_class = 0;
    blank_dialog->name = 0;

    GlobalUnlock(hgbl); 
    
    DialogBoxIndirectParam(GetModuleHandle(0),(LPDLGTEMPLATE)hgbl,parent,proc,(LPARAM)param); 
	
    GlobalFree(hgbl); 
}

int WINAPI WinMain(HINSTANCE hInstance,HINSTANCE hPrevInstance,LPSTR lpCmdLine,int nCmdShow)
{
	MSG msg;
	int ret;
	LOGFONT lf;

#ifdef _DEBUG
	{
		COORD coord;

		AllocConsole();

		coord.X = 256;
		coord.Y = 4096;
		SetConsoleScreenBufferSize(GetStdHandle(STD_OUTPUT_HANDLE),coord);	
	}

	debug_printf(L"New Shit\n");
#endif

	InitCommonControls();

	SystemParametersInfo(SPI_GETICONTITLELOGFONT,sizeof(lf),(PVOID)&lf,0);

	lf.lfWeight = FW_BOLD;
	bold_font = CreateFontIndirect(&lf);

	os_create_blank_dialog(0,_newshit_proc,0);
	
	return 0;
	
}

int main(int,char **)
{
	WinMain(0,0,0,0);
}
